(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: mapred_io.mli 280 2010-10-27 18:13:07Z gerd $ *)

(** Utility library for record-based I/O *)

(** A file consists here of lines. Each line is terminated by LF, and
    is considered as a record.

    The record reader can be used to iterate over a whole file, or
    only a part. For the latter, it is assumed that the file is processed
    bigblock by bigblock. Of course, it is then possible that the lines do not
    end at bigblock boundaries. However, it must be determined whether the
    block reader for bigblock N or the block reader for bigblock N+1 processes
    such lines. The following rules do that:

    - The first bigblock does not have this problem. Its first line is always
      processed by the block reader for block 0.
    - For the other bigblocks we define that the first processed line
      is the line starting after the first LF character in the bigblock.
    - For all bigblocks we define that the last processed line is the
      line following the last LF character in the bigblock. The line can
      be stored partly, even fully in the next bigblock. The block reader
      has to read the next bigblock to allow this.
    - A bigblock must contain at least one LF character (and hence, have
      at least one line to process)

    For best efficiency, the block reader should not be used for
    reading individual bigblocks, but for contiguous ranges of bigblocks.
 *)

class type record_config =
object
  method bigblock_size : int
    (** Same as {!Mapred_def.mapred_job.bigblock_size}. This is the
	requested size. The actual size is rounded up to the next
	multiple of filesystem blocks.
     *)
  method mr_buffer_size : int
    (** Same as {!Mapred_config.mapred_config.mr_buffer_size}. Also
	rounded up.
     *)
  method mr_buffer_size_tight : int
    (** Same as {!Mapred_config.mapred_config.mr_buffer_size_tight}.
	Also rounded up.
     *)
end

class type record_reader =
object
  method pos_in : int
    (** The ordinal number of the record that will be read next. Numbers
	start at 0
     *)

  method input_record : unit -> string
    (** Reads one record, and advances the cursor. May raise [End_of_file] *)

  method peek_record : unit -> string
    (** Reads the record, but does not advance the cursor, i.e. the
	next [peek_record] or [input_record] will read the line again.
	May raise [End_of_file].
     *)

  method close_in : unit -> unit
    (** Releases resources (e.g. closes transactions) *)

  method abort : unit -> unit
    (** Drops resources - intended to be used for error cleanup *)

  method to_fd_e : Unix.file_descr -> unit Uq_engines.engine
    (** [to_fd_e esys fd]: The remaining records are written to [fd].
	The position [pos_in] is not updated. The length of the records
	is not checked except for a few records that are crucial for
	interpreting the boundaries of the bigblocks.

	While the engine is running the other method must not be called.
     *)

  method plasma_cluster : Plasma_client.plasma_cluster
    (** The PlasmaFS cluster *)

end


class type record_writer =
object
  method pos_out : int
    (** The ordinal number of the record that will be written next. Numbers
	start at 0
     *)

  method bpos_out : int64
    (** The byte position of the record that will be written next *)

  method output_record : string -> unit
    (** Outputs these records. The records must not contain LF chars.
	The length is limited to the blocksize-1.
     *)

  method flush : unit -> unit
    (** Flushes the records from the buffer to disk *)

  method from_fd_e : Unix.file_descr -> unit Uq_engines.engine
    (** Outputs the records coming from this file. The file is read until EOF.
	The [pos_out] number is not updated, but [bpos_out] is.
	While the engine is running [output_record] and [flush] must
	not be called.

	The file must be line-structured. If the LF after the last line
	is missing it is silently added. The length of the lines is not
	checked except in a few situations.
     *)

  method close_out : unit -> unit
    (** Releases resources (e.g. closes transactions) *)

  method abort : unit -> unit
    (** Drops resources - intended to be used for error cleanup *)

  method plasma_cluster : Plasma_client.plasma_cluster
    (** The PlasmaFS cluster *)
end

val bigblock_size : Plasma_client.plasma_cluster -> int -> int
  (** [bigblock_size c suggested]: Returns the real bigblock size for
      the [suggested] value. The real size is rounded up to the
      next block multiple.

      The blocksize is retrieved from [c] (synchronously)
   *)


val read_file : Plasma_client.plasma_cluster -> 
                record_config ->
                string -> int64 -> int64 -> record_reader
  (** [read_file c rc name block len]: Reads from [name], starting at [block],
      ending at [block+len-1]. Reading is done in a separate transaction.

      Note that len>=1 is a requirement here. Also, [block] must be
      a multiple of the size of bigblocks.

      The cluster [c] is set to aborted state when not used. Note that this
      also affects all transactions unrelated to [read_file], so is best
      to create a separate [plasma_cluster] object for reading.
   *)

type sync_readers =
    (unit -> record_reader) list

type async_readers =
    (unit -> record_reader Uq_engines.engine) list

val read_multiple : Plasma_client.plasma_cluster -> 
                    record_config ->
                    readers:[`Sync of sync_readers | `Async of async_readers] ->
                    unit -> 
                      record_reader
  (** Constructs a record reader that reads from the input [readers] one
      after the other. The readers can be given in a synchronous form,
      or in an asynchronous form. The latter is preferrable when the
      reader is in asynchronous mode (i.e. when [to_fd_e] is
      running).
   *)

val write_file : Plasma_client.plasma_cluster -> record_config ->
                 string -> record_writer
  (** [write_file c rc name]: Appends records to this file (which must already
      exist). Writing is done in separate transactions.

      As [read_file], the cluster [c] is set to aborted state when not used.
      Note that this also affects all transactions unrelated to [read_file],
      so is best to create a separate [plasma_cluster] object for writing.
   *)

val write_multiple : Plasma_client.plasma_cluster -> record_config ->
                     string -> int64 ->
                     create_sync:(string -> int -> string) ->
                     ?create_async:(string -> int -> string Uq_engines.engine) ->
                     unit ->
                        record_writer
  (** [write_multiple c rc prefix limit create_sync create_async ()]:
      Writes into a sequence of files
      whose names are composed of [prefix] followed by an integer [k]. The
      files are created by calling [create_async prefix k]. A new file is
      started when the current file reaches the size [limit] (in bytes).

      [create_async]: If passed, this function is used instead of [create_sync]
      in asynchronous mode (i.e. when [from_fd_e] is running).
   *)
			 

val create_file_e : ?repl:int -> Plasma_client.plasma_cluster -> string ->
                       unit Uq_engines.engine
val create_file : ?repl:int -> Plasma_client.plasma_cluster -> string -> unit
  (** [create_file c name]: Creates this file exclusively. [repl] is
      the replication factor, 0 by default (i.e. use server default).
   *)


val delete_file :  Plasma_client.plasma_cluster -> string -> unit
  (** Delete this file *)

val file_blocks : Plasma_client.plasma_cluster -> string -> int64
  (** Get the length of the file in blocks *)
