(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: plasma_shm.mli 239 2010-06-23 16:49:03Z gerd $ *)

(** Management of shared memory *)

(** The idea of this module is to count the shared memory allocations,
    and to provide information whether shared memory is well available
    or tight
 *)

type shm_id = [`Shm_id of int ]

type shm_condition = [ `Normal | `Tight ]

class type shm_manager =
object
  method new_id : unit -> shm_id
    (** allocates a new ID for an shm object *)

  method book_allocation : shm_id -> int -> unit
    (** Books this ID with this (total) size *)

  method shm_condition : shm_condition
    (** Estimate where we are *)
end


class null_shm_manager : unit -> shm_manager
  (** This manager always reports the condition [`Normal] *)

class netplex_shm_manager : unit -> shm_manager
  (** This manager puts the shm accouning information into Netplex
      shared variables. This only works from Netplex components.
      Management must be initialized, see the module
      {!Plasma_shm.Netplex_shm_manager}.
   *)

module Netplex_shm_manager : sig
  (** Configuration and initialization of the Netplex shm manager. *)

  val post_add_hook : Netplex_types.controller -> unit
    (** This function must be called from the [post_add_hook] of the
	component
     *)

  val init : low:int64 -> high:int64 -> unit
    (** This can be called from component context to set the low-water
	and high-water marks. By default, these marks are [max_int],
	and [shm_condition] always returns [`Normal].

	When the sum of all allocated shared memory exceeds the high-water
	mark, the condition is switched to [`Tight]. It is switched back
	to [`Normal] when the sum sinks below the low-water mark.
     *)
end


